from pydantic import Field
from fakts.discovery.base import Discovery
from fakts.discovery.static import StaticDiscovery
from fakts.discovery.base import FaktsEndpoint
from fakts.grants.base import FaktsGrant
import ssl
import certifi
from pydantic import BaseModel
import aiohttp
from typing import Any, Dict
from .errors import ClaimError
import logging

logger = logging.getLogger(__name__)


Token = str
EndpointUrl = str


class FaktClaim(BaseModel):
    """FaktClaim

    A FaktClaim is a claim for a Fakt. It is used to claim a Fakt on a Fakts endpoint.

    """

    version: str
    identifier: str
    endpoint_url: str
    token: str


class CacheFile(BaseModel):
    """Cache file model"""

    claims: Dict[EndpointUrl, FaktClaim]


class RemoteGrant(FaktsGrant):
    """Abstract base class for remote grants

    A Remote grant is a grant that connects to a fakts server,
    and tires to establishes a secure relationship with it.

    This is done by providing the fakts server with a software
    manifest consisting of a world unique identifier, and a
    version number.

    The fakts server then can depending on the grant type
    respond with a token that then in turn can be used to
    retrieve the configuration from the fakts server.

    """

    version: str = Field(
        None,
        description="The version of this app, if we are using a remote grant, this will be used to identify the app",
    )
    """ The version hint of this app. Will be used to uniquely identify this app on the endpoint"""

    identifier: str = Field(None, description="The unique identifier of the app")
    """ A world unique identifier of this app. This is used to identify the app on the fakts server."""

    discovery: Discovery = Field(default_factory=StaticDiscovery)
    "The discovery method to use, if not specified, the static discovery will be used"

    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    auto_demand_on_failure: bool = True
    """ If set to true, the grant will try to demand a new token if the claim fails"""

    force_refresh: bool = False
    """Should we always force a refresh of the token. If we have no cached it?"""

    async def aload(self, force_refresh: bool = False):
        """Load the configuration from the remote endpoint

        This function will load the configuration from the remote endpoint.
        It will first try to load the configuration from the cache file.
        If this fails, it will try to load the configuration from the endpoint.
        If this fails, it will try to demand a new token from the endpoint.
        If this fails, it will raise an exception.

        """
        endpoint = await self.discovery.discover()
        token = await self.ademand(endpoint)

        return await self.aclaim(token, endpoint)

    async def ademand(self, endpoint: FaktsEndpoint) -> Token:
        """Demand a token for receiving the configuration, for this
        specific app"""

        raise NotImplementedError(
            "This is an abstract base Class. Please use one of the subclasses"
        )

    async def aclaim(self, token: Token, endpoint: FaktsEndpoint) -> Dict[str, Any]:
        """Claim the configuration from the endpoint"""

        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context)
        ) as session:
            async with session.post(
                f"{endpoint.base_url}claim/",
                json={
                    "token": token,
                },
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if not "status" in data:
                        raise ClaimError("Malformed Answer")

                    status = data["status"]
                    if status == "error":
                        raise ClaimError(data["message"])
                    if status == "granted":
                        return data["config"]

                    raise ClaimError(f"Unexpected status: {status}")
                else:
                    raise Exception("Error! Coud not claim this app on this endpoint")

    class Config:
        arbitrary_types_allowed = True



from pydantic import Field
from fakts.grants.base import FaktsGrant
from fakts.grants.errors import GrantError
import yaml
from koil.qt import QtCoro, QtFuture
from qtpy import QtWidgets


class NoFileSelected(GrantError):
    pass


class QtSelectYaml(QtWidgets.QFileDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFileMode(QtWidgets.QFileDialog.ExistingFile)
        self.setNameFilter("YAML files (*.yaml)")

    @classmethod
    def ask(self, parent=None):
        filepath, weird = self.getOpenFileName(parent=parent, caption="Select a Yaml")
        return filepath


class WrappingWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.get_file_coro = QtCoro(self.open_file)

    def open_file(self, future: QtFuture):
        filepath = QtSelectYaml.ask(parent=self)

        if filepath:
            future.resolve(filepath)
        else:
            future.reject(NoFileSelected("No file selected"))


class QtYamlGrant(FaktsGrant):
    """A grant that allows the user to select a yaml file.
    on the Ui. The yaml file is then loaded and returned as a dict."""

    widget: WrappingWidget = Field(exclude=True)

    async def aload(self, **kwargs):
        filepath = await self.widget.get_file_coro.acall()
        with open(filepath, "r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

        return config

    class Config:
        arbitrary_types_allowed = True
        json_encoders = {QtWidgets.QFileDialog: lambda x: x.__class__.__name__}



from fakts.grants.base import FaktsGrant
import toml


class TomlGrant(FaktsGrant):
    filepath: str

    async def aload(self, force_refresh=False, **kwargs):
        with open(self.filepath, "r") as file:
            config = toml.load(file)

        return config



from socket import socket
from fakts.discovery.base import Discovery
from fakts.discovery.base import Beacon, FaktsEndpoint
from typing import Dict, Optional, AsyncGenerator

from pydantic import Field
from socket import socket, AF_INET, SOCK_DGRAM
import asyncio
import json
from koil import unkoil
import logging
import os
import yaml
import pydantic
from pydantic import BaseModel
import aiohttp
import ssl
import certifi
from .utils import discover_url
from typing import Optional, List

logger = logging.getLogger(__name__)


class DiscoveryProtocol(asyncio.DatagramProtocol):
    pass

    def __init__(self, recvq) -> None:
        super().__init__()
        self._recvq = recvq

    def datagram_received(self, data, addr):
        self._recvq.put_nowait((data, addr))


class AdvertisedConfig(BaseModel):
    selected_endpoint: Optional[FaktsEndpoint]


class ListenBinding(BaseModel):
    address: str = ""
    port: int = 45678
    magic_phrase: str = "beacon-fakts"


async def alisten(bind: ListenBinding, strict: bool = False) -> AsyncGenerator[Beacon, None]:

    s = socket(AF_INET, SOCK_DGRAM)  # create UDP socket
    s.bind((bind.address, bind.port))

    try:

        loop = asyncio.get_event_loop()
        read_queue = asyncio.Queue()
        transport, pr = await loop.create_datagram_endpoint(
            lambda: DiscoveryProtocol(read_queue), sock=s
        )

        while True:
            data, addr = await read_queue.get()
            try:
                data = str(data, "utf8")
                if data.startswith(bind.magic_phrase):
                    endpoint = data[len(bind.magic_phrase) :]

                    try:
                        endpoint = json.loads(endpoint)
                        endpoint = Beacon(**endpoint)
                        yield endpoint

                    except json.JSONDecodeError as e:

                        logger.error("Received Request but it was not valid json")
                        if strict:
                            raise e

                else:
                    logger.error(
                        f"Received Non Magic Response {data}. Maybe somebody sends"
                    )

            except UnicodeEncodeError as e:
                logger.error("Couldn't decode received message")
                if strict:
                    raise e

    except asyncio.CancelledError as e:
        transport.close()
        s.close()
        logger.info("Stopped checking")
        raise e
    finally:
        transport.close()
        s.close()
        logger.info("Stopped checking")


async def alisten_pure(bind: ListenBinding, strict: bool = False) -> AsyncGenerator[Beacon, None]:

    already_detected = set()

    async for x in alisten(bind, strict):
        if x.url not in already_detected:
            already_detected.add(x.url)
            yield x



class AdvertisedDiscovery(Discovery):
    broadcast_port = 45678
    magic_phrase = "beacon-fakts"
    bind = ""
    strict: bool = False
    discovered_endpoints: Dict[str, FaktsEndpoint] = Field(default_factory=dict)
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""
    allow_appending_slash: bool = Field(
        default=True,
        description="If the url does not end with a slash, should we append one? ",
    )
    auto_protocols: List[str] = Field(
        default_factory=lambda: [],
        description="If no protocol is specified, we will try to connect to the following protocols",
    )
    timeout: int = Field(
        default=3,
        description="The timeout for the connection",
    )

    async def discover(self, force_refresh=False):
        if os.path.exists(self.file):
            with open(self.file, "r") as f:
                x = yaml.load(f, Loader=yaml.FullLoader)
                try:
                    cache = AdvertisedConfig(**x)
                except pydantic.ValidationError as e:
                    logger.error(f"Could not load cache file: {e}. Ignoring it")
                    cache = AdvertisedConfig()
        else:
            cache = AdvertisedConfig()

        if not cache.selected_endpoint or force_refresh:
            binding = ListenBinding(address=self.bind, port=self.broadcast_port, magic_phrase=self.magic_phrase)
            async for beacon in alisten_pure(binding, strict=self.strict):
                try:
                    endpoint = await discover_url(beacon, self.ssl_context)
                    return endpoint
                except Exception as e:
                    logger.error(f"Could not connect to beacon {beacon.url}: {e}")
                    continue



    class Config:
        arbitrary_types_allowed = True